
from .structs import *
from .type import *


class Method:
    def __init__(self, library, objc_class, method: objc2_meth, vmaddr: int):
        self.objc_class = objc_class

        try:
            self.sel = library.get_cstr_at(method.selector, 0, vm=True, sectname="__objc_methname")
            typestr = library.get_cstr_at(method.types, 0, vm=True, sectname="__objc_methtype")
        except ValueError as ex:
            selref = library.get_bytes(method.selector+vmaddr, 8, vm=True)
            self.sel = library.get_cstr_at(selref, 0, vm=True, sectname="__objc_methname")
            typestr = library.get_cstr_at(method.types+vmaddr+4, 0, vm=True, sectname="__objc_methtype")
        except Exception as ex:
            raise ex

        self.typestr = typestr
        self.types = library.tp.process(typestr)
        if len(self.types) == 0:
            raise ValueError("Empty Typestr")

        self.return_string = self._renderable_type(self.types[0])
        self.arguments = [self._renderable_type(i) for i in self.types[1:]]

        self.signature = self._build_method_signature()

    def __str__(self):
        ret = ""
        ret += self.signature
        return ret

    @staticmethod
    def _renderable_type(type: Type):
        if type.type == EncodedType.NORMAL:
            return str(type)
        elif type.type == EncodedType.STRUCT:
            ptraddon = ""
            for i in range(0, type.pointer_count):
                ptraddon += '*'
            return ptraddon + type.value.name

    def _build_method_signature(self):
        dash = "+" if self.objc_class.meta else "-"
        ret = "(" + self.return_string + ")"

        if len(self.arguments) == 0:
            return dash + ret + self.sel

        segs = []
        for i, item in enumerate(self.sel.split(':')):
            if item == "":
                continue
            try:
                segs.append(item + ':' + '(' + self.arguments[i+2] + ')' + 'arg' + str(i) + ' ')
            except IndexError:
                segs.append(item)

        sig = ''.join(segs)

        return dash + ret + sig



from enum import IntEnum

from collections import namedtuple

action = namedtuple("action", ["vmaddr", "libname", "item"])
record = namedtuple("record",
                    ["seg_index", "seg_offset", "lib_ordinal", "type", "flags", "name", "addend", "special_dylib"])


class BindingProcessor:
    """
    This doesn't do a whole lot at the moment;

    It simply parses through the binding info in the library, and then creates a list of actions specified in the
        binding info.
    """

    def __init__(self, lib):
        self.lib = lib
        self.import_stack = self._load_binding_info()
        self.actions = self._create_action_list()

    def _create_action_list(self):
        actions = []
        for bind_command in self.import_stack:
            segment = list(self.lib.segments.values())[bind_command.seg_index]
            vm_addr = segment.vmaddr + bind_command.seg_offset
            try:
                lib = self.lib.linked[bind_command.lib_ordinal - 1].install_name
            except IndexError:
                lib = str(bind_command.lib_ordinal)
            item = bind_command.name
            actions.append(action(vm_addr & 0xFFFFFFFFF, lib, item))
        return actions

    def _load_binding_info(self):
        lib = self.lib
        ea = lib.info.bind_off
        import_stack = []
        while True:
            # print(hex(ea))
            if ea - lib.info.bind_size >= lib.info.bind_off:
                break
            seg_index = 0x0
            seg_offset = 0x0
            lib_ordinal = 0x0
            btype = 0x0
            flags = 0x0
            name = ""
            addend = 0x0
            special_dylib = 0x1
            while True:
                # There are 0xc opcodes total
                # Bitmask opcode byte with 0xF0 to get opcode, 0xF to get value
                OP = self.lib.get_bytes(ea, 1) & 0xF0
                VALUE = self.lib.get_bytes(ea, 1) & 0x0F
                ea += 1
                if OP == OPCODE.BIND_OPCODE_DONE:
                    import_stack.append(
                        record(seg_index, seg_offset, lib_ordinal, btype, flags, name, addend, special_dylib))
                    break
                elif OP == OPCODE.BIND_OPCODE_SET_DYLIB_ORDINAL_IMM:
                    lib_ordinal = VALUE
                elif OP == OPCODE.BIND_OPCODE_SET_DYLIB_ORDINAL_ULEB:
                    lib_ordinal, bump = self.lib.decode_uleb128(ea)
                    ea = bump
                elif OP == OPCODE.BIND_OPCODE_SET_DYLIB_SPECIAL_IMM:
                    special_dylib = 0x1
                    lib_ordinal = VALUE
                elif OP == OPCODE.BIND_OPCODE_SET_SYMBOL_TRAILING_FLAGS_IMM:
                    flags = VALUE
                    name = self.lib.get_cstr_at(ea)
                    ea += len(name)
                    ea += 1
                elif OP == OPCODE.BIND_OPCODE_SET_TYPE_IMM:
                    btype = VALUE
                elif OP == OPCODE.BIND_OPCODE_SET_ADDEND_SLEB:
                    ea += 1
                elif OP == OPCODE.BIND_OPCODE_SET_SEGMENT_AND_OFFSET_ULEB:
                    seg_index = VALUE
                    number, head = self.lib.decode_uleb128(ea)
                    seg_offset = number
                    ea = head
                elif OP == OPCODE.BIND_OPCODE_ADD_ADDR_ULEB:
                    o, bump = self.lib.decode_uleb128(ea)
                    seg_offset += o
                    ea = bump
                elif OP == OPCODE.BIND_OPCODE_DO_BIND_ADD_ADDR_ULEB:
                    import_stack.append(
                        record(seg_index, seg_offset, lib_ordinal, btype, flags, name, addend, special_dylib))
                    seg_offset += 8
                    o, bump = self.lib.decode_uleb128(ea)
                    seg_offset += o
                    ea = bump

                elif OP == OPCODE.BIND_OPCODE_DO_BIND_ADD_ADDR_IMM_SCALED:
                    import_stack.append(
                        record(seg_index, seg_offset, lib_ordinal, btype, flags, name, addend, special_dylib))
                    seg_offset = seg_offset + (VALUE * 8) + 8
                elif OP == OPCODE.BIND_OPCODE_DO_BIND_ULEB_TIMES_SKIPPING_ULEB:
                    t, bump = self.lib.decode_uleb128(ea)
                    count = t
                    ea = bump
                    s, bump = self.lib.decode_uleb128(ea)
                    skip = s
                    ea = bump
                    for i in range(0, count):
                        import_stack.append(
                            record(seg_index, seg_offset, lib_ordinal, btype, flags, name, addend, special_dylib))
                        seg_offset += skip + 8
                elif OP == OPCODE.BIND_OPCODE_DO_BIND:
                    import_stack.append(
                        record(seg_index, seg_offset, lib_ordinal, btype, flags, name, addend, special_dylib))
                    seg_offset += 8
                else:
                    assert 0 == 1

        return import_stack


class OPCODE(IntEnum):
    BIND_OPCODE_DONE = 0x0
    BIND_OPCODE_SET_DYLIB_ORDINAL_IMM = 0x10
    BIND_OPCODE_SET_DYLIB_ORDINAL_ULEB = 0x20
    BIND_OPCODE_SET_DYLIB_SPECIAL_IMM = 0x30
    BIND_OPCODE_SET_SYMBOL_TRAILING_FLAGS_IMM = 0x40
    BIND_OPCODE_SET_TYPE_IMM = 0x50
    BIND_OPCODE_SET_ADDEND_SLEB = 0x60
    BIND_OPCODE_SET_SEGMENT_AND_OFFSET_ULEB = 0x70
    BIND_OPCODE_ADD_ADDR_ULEB = 0x80
    BIND_OPCODE_DO_BIND = 0x90
    BIND_OPCODE_DO_BIND_ADD_ADDR_ULEB = 0xa0
    BIND_OPCODE_DO_BIND_ADD_ADDR_IMM_SCALED = 0xb0
    BIND_OPCODE_DO_BIND_ULEB_TIMES_SKIPPING_ULEB = 0xc0



__all__ = ['Segment', 'Section']

from .structs import *

class Segment:
    """


    segment_command_64:
    ["off", "cmd", "cmdsize", "segname", "vmaddr", "vmsize", "fileoff", "filesize",
        "maxprot", "initprot", "nsects", "flags"]
    """
    def __init__(self, library, cmd):
        self.library = library
        self.cmd = cmd
        self.vmaddr = cmd.vmaddr
        self.fileaddr = cmd.fileoff
        self.size = cmd.vmsize
        self.name = ""
        for i, c in enumerate(hex(cmd.segname)[2:]):
            if i % 2 == 0:
                self.name += chr(int(c + hex(cmd.segname)[2:][i+1], 16))
        self.name = self.name[::-1]
        self.sections = self._process_sections()

    def _process_sections(self):
        sections = {}
        ea = self.cmd.off + sizeof(segment_command_64_t)

        for section in range(0, self.cmd.nsects):
            sect = self.library.load_struct(ea, section_64_t)
            section = Section(self, sect)
            sections[section.name] = section
            ea += sizeof(section_64_t)

        ea += sizeof(segment_command_64_t)
        return sections


class Section:
    """


    section_64
     ["off",
     "sectname",
     "segname",
     "addr", VM ADDRESS
     "size", SIZE ON BOTH
     "offset", FILE ADDRESS
     "align", "reloff", "nreloc", "flags", "void1", "void2", "void3"]
    """
    def __init__(self, segment, cmd):
        self.cmd = cmd
        self.segment = segment
        self.name = segment.library.get_str_at(cmd.off, 16)
        self.vmaddr = cmd.addr
        self.fileaddr = cmd.offset
        self.size = cmd.size




from collections import namedtuple

"""
C Struct representation in Python
I have done what I can to make the syntax used similar to C syntax

The `struct` variable is a subclass of Tuple which contains a struct 'typedef' and a list of byte sizes for each field in the struct, in order.
"""
struct = namedtuple("struct", ["struct", "sizes"])

"""
OBJC STRUCTS
VALUES OF FIELDS SHOULD LINE UP WITH THOSE IN OBJC4 SOURCE
"""
objc2_class = namedtuple("objc2_class", ["off", "isa", "superclass", "cache", "vtable", "info"])
objc2_class_t = struct(objc2_class, [8, 8, 8, 8, 8])
objc2_class_ro = namedtuple("objc2_class_ro", ["off", "flags", "ivar_base_start", "ivar_base_size", "reserved", "ivar_lyt", "name", "base_meths", "base_prots", "ivars", "weak_ivar_lyt", "base_props"])
objc2_class_ro_t = struct(objc2_class_ro, [4, 4, 4, 4, 8, 8, 8, 8, 8, 8, 8])

objc2_meth = namedtuple("objc2_meth", ["off", "selector", "types", "imp"])
objc2_meth_t = struct(objc2_meth, [8, 8, 8])
objc2_meth_list_entry_t = struct(objc2_meth, [4,4,4])
objc2_meth_list = namedtuple("objc2_meth_list", ["off", "entrysize", "count"])
objc2_meth_list_t = struct(objc2_meth_list, [4, 4])

objc2_prop_list = namedtuple("objc2_prop_list", ["off", "entrysize", "count"])
objc2_prop_list_t = struct(objc2_prop_list, [4, 4])

objc2_prop = namedtuple("objc2_prop", ["off", "name", "attr"])
objc2_prop_t = struct(objc2_prop, [8, 8])

objc2_prot_list = namedtuple("objc2_prot_list", ["off", "cnt"])
objc2_prot_list_t = struct(objc2_prot_list, [8])

objc2_prot = namedtuple("objc2_prot", ["off", "isa", "name", "prots", "inst_meths", "class_meths", "opt_inst_meths", "opt_class_meths", "inst_props", "cb", "flags"])
objc2_prot_t = struct(objc2_prot, [8, 8, 8, 8, 8, 8, 8, 8, 4, 4])

objc2_ivar_list = namedtuple("objc2_ivar_list", ["off", "entrysize", "cnt"])
objc2_ivar_list_t = struct(objc2_ivar_list, [4, 4])

objc2_ivar = namedtuple("objc2_ivar", ["off", "offs", "name", "type", "align", "size"])
objc2_ivar_t = struct(objc2_ivar, [8, 8, 8, 4, 4])


def sizeof(t: struct):
    assert isinstance(t, struct)
    return sum(t.sizes)


from macho.library import Library

from .structs import *
from .type import TypeProcessor

from .objcclass import Class
from macho.segment import Segment, Section


class ObjCLibrary:
    def __init__(self, library):
        self.library = library
        self.tp = TypeProcessor()
        self.name = library.name

        self.classlist = self._generate_classlist(None)

    def _generate_classlist(self, classlimit):
        sect: Section = self.library.segments['__DATA_CONST'].sections['__objc_classlist']
        classes = []
        cnt = sect.size // 0x8
        for i in range(0, cnt):
            if classlimit is None:
                classes.append(Class(self, sect.vmaddr + i * 0x8))
            else:
                oc = Class(self, sect.vmaddr + i * 0x8)
                if classlimit == oc.name:
                    classes.append(oc)
        return classes

    def get_bytes(self, offset: int, length: int, vm=False, sectname=None):
        return self.library.get_bytes(offset, length, vm, sectname)

    def load_struct(self, addr: int, struct_type: struct, vm=True, sectname=None, endian="little"):
        return self.library.load_struct(addr, struct_type, vm, sectname, endian)

    def get_str_at(self, addr: int, count: int, vm=True, sectname=None):
        return self.library.get_str_at(addr, count, vm, sectname)

    def get_cstr_at(self, addr: int, limit: int = 0, vm=True, sectname=None):
        return self.library.get_cstr_at(addr, limit, vm, sectname)




# Cell
from collections import namedtuple

from .segment import Segment, Section

# Cell
vm_obj = namedtuple("vm_obj", ["vmaddr", "vmend", "size", "fileaddr", "name"])


class _VirtualMemoryMap:
    """
    Virtual Memory is the location "in memory" where the library/bin, etc will be accessed when ran
    This is not where it actually sits in memory at runtime; it will be slid, but the program doesnt know and doesnt care
    The slid address doesnt matter to us either, we only care about the addresses the rest of the file cares about

    There are two address sets used in mach-o files: vm, and file. (commonly; vmoff and fileoff)
    For example, when reading raw data of an executable binary:
    0x0 file offset will (normally?) map to 0x10000000 in the VM

    These VM offsets are relayed to the linker via Load Commands
    Some locations in the file do not have VM counterparts (examples being symbol table(citation needed))

    Some other VM related offsets are changed/modified via binding info(citation needed)
    """

    def __init__(self, slice):
        # name: vm_obj
        self.slice = slice
        self.map = {}
        self.stats = {}

    def __str__(self):
        """
        We want to be able to just call print(macho.vm) to display the filemap in a human-readable format

        :return: multiline String representation of the filemap
        """

        ret = ""
        # Sort our map by VM Address, this should already be how it is but cant hurt
        sortedmap = {k: v for k, v in sorted(self.map.items(), key=lambda item: item[1].vmaddr)}

        for (key, obj) in sortedmap.items():
            # 'string'.ljust(16) adds space padding
            # 'string'[2:].zfill(9) removes the first 2 chars and pads the string with 9 zeros
            #       then we just re-add the 0x manually.

            # this gives us a nice list with visually clear columns and rows
            ret += f'{key.ljust(16)}  ||  Start: 0x{hex(obj.vmaddr)[2:].zfill(9)}  |  End: 0x{hex(obj.vmend)[2:].zfill(9)}  |  Size: 0x{hex(obj.size)[2:].zfill(9)}  |  Slice ' \
                   f'Offset:  0x{hex(obj.fileaddr)[2:].zfill(9)}  ||  File Offset: 0x{hex(obj.fileaddr + self.slice.offset)[2:].zfill(9)}\n'
        return ret

    def get_vm_start(self):
        """
        Get the address the VM starts in, excluding __PAGEZERO
        Method selector dumping uses this
        :return:
        """
        sortedmap = {k: v for k, v in sorted(self.map.items(), key=lambda item: item[1].vmaddr)}
        if list(sortedmap.keys())[0] == "__PAGEZERO":
            return list(sortedmap.values())[1].vmaddr
        else:
            return list(sortedmap.values())[0].vmaddr

    def get_file_address(self, vm_address: int, sname = None):
        # This function gets called hundreds of thousands of times during processing and is the main source of overhead
        if sname is not None:
            o = self.map[sname]
            if vm_address >= o.vmaddr and o.vmend >= vm_address:
                return o.fileaddr + vm_address - o.vmaddr
            else:
                try:
                    o = self.map['__EXTRA_OBJC']
                    if vm_address >= o.vmaddr and o.vmend >= vm_address:
                        return o.fileaddr + vm_address - o.vmaddr
                except:
                    for o in self.map.values():
                        if vm_address >= o.vmaddr and o.vmend >= vm_address:
                            # self.stats[o.name] += 1
                            return o.fileaddr + vm_address - o.vmaddr
        for o in self.map.values():
            if vm_address >= o.vmaddr and o.vmend >= vm_address:
                #self.stats[o.name] += 1
                return o.fileaddr + vm_address - o.vmaddr
        #print(f'VM MAPPING:')
        #print(self)
        raise ValueError(f'Address {hex(vm_address)} couldn\'t be found in vm address set')

    def add_segment(self, segment: Segment):
        if len(segment.sections) == 0:
            seg_obj = vm_obj(segment.vmaddr, segment.vmaddr+segment.size, segment.size, segment.fileaddr, segment.name)
            self.map[segment.name] = seg_obj
            self.stats[segment.name] = 0
        else:
            for section in segment.sections.values():
                name = section.name if section.name not in self.map.keys() else section.name + '2'
                sect_obj = vm_obj(section.vmaddr, section.vmaddr+section.size, section.size, section.fileaddr, name)
                self.map[name] = sect_obj
                self.smap = {k: v for k, v in sorted(self.map.items(), key=lambda item: item[1].vmaddr)}
                self.stats[name] = 0









from enum import Enum

type_encodings = {
    "c": "char",
    "i": "int",
    "s": "short",
    "l": "long",
    "q": "NSInteger",
    "C": "unsigned char",
    "I": "unsigned int",
    "S": "unsigned short",
    "L": "unsigned long",
    "Q": "NSUInteger",
    "f": "float",
    "d": "CGFloat",
    "b": "BOOL",
    "@": "id",
    "B": "bool",
    "v": "void",
    "*": "char *",
    "#": "Class",
    ":": "SEL",
    "?": "unk",
}


class Struct:
    def __init__(self, processor, type_str: str):
        # {name=dd}

        # Remove the outer {}, then get everything to the left of the equal sign
        self.name = type_str[1:-1].split('=')[0]

        # Remove the outer {},
        # get everything after the first = sign,
        # Process that via the processor
        # Save the resulting list to self.fields
        self.fields = processor.process(type_str[1:-1].split('=', 1)[1])

    def __str__(self):
        ret = "typedef struct " + self.name + " {\n"
        for i, field in enumerate(self.fields):
            if isinstance(field.value, Struct):
                field = field.value.name
            else:
                field = field.value
            ret += "    " + field + ' field' + str(i) + ';\n'
        ret += '} ' + self.name + ';'
        if len(self.fields) == 0:
            ret += " // Empty Struct"
        return ret


class EncodingType(Enum):
    METHOD = 0
    PROPERTY = 1
    IVAR = 2


class EncodedType(Enum):
    STRUCT = 0
    NAMED = 1
    ID = 2
    NORMAL = 3


class Type:
    def __init__(self, processor, typestr, pc=0):
        start = typestr[0]
        self.child = None
        self.pointer_count = pc

        if start in type_encodings.keys():
            self.type = EncodedType.NORMAL
            self.value = type_encodings[start]
            return

        elif start == '"':
            self.type = EncodedType.NAMED
            self.value = typestr[1:-1]
            return

        elif start == '{':
            self.type = EncodedType.STRUCT
            self.value = Struct(processor, typestr)
            return
        raise ValueError(f'Struct with type {start} not found')

    def __str__(self):
        pref = ""
        for i in range(0, self.pointer_count):
            pref += "*"
        return pref + str(self.value)


class TypeProcessor:
    def __init__(self):
        self.structs = {}

    def save_struct(self, struct: Struct):
        if struct.name not in self.structs.keys():
            self.structs[struct.name] = struct

    def process(self, type: str):
        try:
            tokens = self.tokenize(type)
            types = []
            pc = 0
            for i, token in enumerate(tokens):
                if token == "^":
                    pc += 1
                else:
                    typee = Type(self, token, pc)
                    types.append(typee)
                    if typee.type == EncodedType.STRUCT:
                        self.save_struct(typee.value)
                    pc = 0
            return types
        except:
            raise AssertionError(type)

    @staticmethod
    def tokenize(type: str):
        # ^Idd^{structZero=dd{structName={innerStructName=dd}}{structName2=dd}}

        # This took way too long to write
        # Apologies for lack of readability, it splits up the string into a list
        # Makes every character a token, except root structs
        #   which it compiles into a full string with the contents and tacks onto said list
        toks = []
        b = False
        bc = 0
        bu = ""
        for c in type:
            if b:
                bu += c
                if c == "{":
                    bc += 1
                elif c == "}":
                    bc -= 1
                    if bc == 0:
                        toks.append(bu)
                        b = False
                        bu = ""
            elif c in type_encodings or c == "^":
                toks.append(c)
            elif c == "{":
                bu += "{"
                b = True
                bc += 1
            elif c == '"':
                try:
                    toks = [type.split('@', 1)[1]]
                except:
                    # Named fields ;_;
                    return []
                break
        return toks




from enum import Enum

from macho.segment import Segment

from macho.binding import BindingProcessor

from macho._vm import _VirtualMemoryMap
from .structs import *
from collections import namedtuple


os_version = namedtuple("os_version", ["x", "y", "z"])


class PlatformType(Enum):
    MACOS = 1
    IOS = 2
    TVOS = 3
    WATCHOS = 4
    BRIDGEOS = 5
    MACCATALYST = 6
    IOSSIMULATOR = 7
    TVOSSIMULATOR = 8
    WATCHOSSIMULATOR = 9
    DRIVERKIT = 10


class ToolType(Enum):
    CLANG = 1
    SWIFT = 2
    LD = 3


class Dylib:
    """
    Represents any dynamiclibrary including self and off-image ones


    """
    def __init__(self, library, cmd):
        self.library = library
        self.install_name = self._get_name(cmd)
        self.local = cmd.cmd == 0xD

    def _get_name(self, cmd):
        ea = cmd.off + sizeof(dylib_command_t)
        return self.library.get_cstr_at(ea)


class LibraryHeader:
    """
    This class represents the Mach-O Header
    It contains the basic header info along with all load commands within it.

    It doesn't handle complex abstraction logic, it simply loads in the load commands as their raw structs
    """
    def __init__(self, slice):
        offset = 0
        self.dyld_header: dyld_header = slice.load_struct(offset, dyld_header_t)
        self.load_commands = []
        self._process_load_commands(slice)

    def _process_load_commands(self, slice):
        """
        This function takes the raw file and parses through its load commands

        :param fd: file
        :return:
        """

        # Start address of the load commands.
        ea = self.dyld_header.off + 0x20

        # Loop through the dyld_header by load command count
        # possibly this could be modified to check for other load commands
        #       as a rare obfuscation technique involves fucking with these to screw with RE tools.

        for i in range(1, self.dyld_header.loadcnt):
            cmd = slice.get_at(ea, 4)
            try:
                load_cmd = slice.load_struct(ea, LOAD_COMMAND_TYPEMAP[cmd])
            except KeyError as ex:
                unk_lc = slice.load_struct(ea, unk_command_t)
                load_cmd = unk_lc

            self.load_commands.append(load_cmd)
            ea += load_cmd.cmdsize


class Library:
    """
    This class represents the Mach-O Binary as a whole.

    It's the root object in the massive tree of information we're going to build up about the binary

    This is an abstracted version, other classes will handle the raw struct interaction;
        here, we facilitate that interaction within those classes and generate our abstract representation

    Calling __init__ on this class will kickstart the full process.
    """
    def __init__(self, slice):
        self.macho_header = LibraryHeader(slice)
        self.slice = slice

        self.linked = []
        self.segments = {}

        self.vm = _VirtualMemoryMap(slice)

        self.info = None
        self.dylib = None
        self.uuid = None

        self.platform = None

        self.minos = None
        self.sdk_version = None


        self._parse_load_commands()
        if self.info:
            self.binding_actions = self._load_binding()

        if self.dylib is not None:
            self.name = self.dylib.install_name.split('/')[-1]
        else:
            self.name = ""

    def get_bytes(self, offset: int, length: int, vm=False, sectname=None):
        if vm:
            offset = self.vm.get_file_address(offset, sectname)
        return self.slice.get_at(offset, length)

    def load_struct(self, addr: int, struct_type: struct, vm=False, sectname=None, endian="little"):
        if vm:
            addr = self.vm.get_file_address(addr, sectname)
        return self.slice.load_struct(addr, struct_type, endian)

    def get_str_at(self, addr: int, count: int, vm=False, sectname=None):
        if vm:
            addr = self.vm.get_file_address(addr, sectname)
        return self.slice.get_str_at(addr, count)

    def get_cstr_at(self, addr: int, limit: int = 0, vm=False, sectname=None):
        if vm:
            addr = self.vm.get_file_address(addr, sectname)
        return self.slice.get_cstr_at(addr, limit)

    def decode_uleb128(self, readHead: int):
        return self.slice.decode_uleb128(readHead)

    def _load_binding(self):
        binding  = BindingProcessor(self)

        return binding.actions

    def _parse_load_commands(self):
        for cmd in self.macho_header.load_commands:
            # my structLoad function *ALWAYS* saves the offset on-disk to the .off field, regardless of the struct
            #   loaded.
            ea = cmd.off

            if isinstance(cmd, segment_command_64):
                segment = Segment(self, cmd)
                self.vm.add_segment(segment)
                self.segments[segment.name] = segment

            if isinstance(cmd, dyld_info_command):
                self.info = cmd

            if isinstance(cmd, uuid_command):
                self.uuid = cmd.uuid

            # https://www.rubydoc.info/gems/ruby-macho/0.1.8/MachO/SourceVersionCommand

            if isinstance(cmd, build_version_command):
                self.platform = PlatformType(cmd.platform)
                self.minos = os_version(x=self.get_bytes(cmd.off + 14, 2), y=self.get_bytes(cmd.off + 13, 1), z=self.get_bytes(cmd.off + 12, 1))
                self.sdk_version = os_version(x=self.get_bytes(cmd.off + 18, 2), y=self.get_bytes(cmd.off + 17, 1), z=self.get_bytes(cmd.off + 16, 1))

            if isinstance(cmd, dylib_command):
                ea += sizeof(dylib_command_t)
                if cmd.cmd == 0xD:  # local
                    self.dylib = Dylib(self, cmd)
                else:
                    self.linked.append(Dylib(self, cmd))




from enum import Enum
from typing import Tuple

from .library import Library
from .structs import *


class CPUType(Enum):
    X86 = 0
    X86_64 = 1
    ARM = 2
    ARM64 = 3


class CPUSubType(Enum):
    X86_64_ALL = 0
    X86_64_H = 1
    ARMV7 = 2
    ARMV7S = 3
    ARM64_ALL = 4
    ARM64_V8 = 5


class Slice:
    def __init__(self, macho_file, arch_struct: fat_arch, offset=0):
        self.macho_file = macho_file
        self.arch_struct = arch_struct

        if self.arch_struct:
            self.offset = arch_struct.offset
            self.type = self._load_type()
            self.subtype = self._load_subtype()
        else:
            self.offset = offset

        self.library = self.load_library()
    def load_library(self):
        return Library(self)

    def load_struct(self, addr: int, struct_type: struct, endian="little"):
        sizeOf = sum(struct_type.sizes)
        fieldNames = list(struct_type.struct.__dict__['_fields'])  # unimportant?
        fields = [addr]
        ea = addr

        for field in struct_type.sizes:
            field_addr = ea
            fields.append(self.get_at(field_addr, field, endian))
            ea += field

        if len(fields) != len(fieldNames):
            raise ValueError(
                f'Field-Fieldname count mismatch in load_struct for {struct.struct.__doc__}.\nCheck Fields and Size Array.')
        return struct_type.struct._make(fields)

    def get_at(self, addr, count, endian="little"):
        addr = addr + self.offset
        return int.from_bytes(self.macho_file.file[addr:addr + count], endian)

    def get_str_at(self, addr: int, count: int):
        addr = addr + self.offset
        return self.macho_file.file[addr:addr + count].decode().rstrip('\x00')

    def get_cstr_at(self, addr: int, limit: int = 0):
        addr = addr + self.offset
        ret = ""
        ea: int = addr
        try:
            while True:
                if limit != 0:
                    if ea - addr >= limit:
                        break
                if ea - addr >= 20000:
                    print(f'Endless String fallback; addr={addr}')
                    print(ret)
                    raise ValueError("Endless String Possibly Detected")
                char = self.macho_file.file[ea:ea + 1].decode()
                if char == '\x00':
                    if len(ret) > 0:
                        break
                    else:
                        ea += 1
                else:
                    ret = ret + char
                    ea += 1
        except UnicodeError as ex:
            raise UnicodeError(f'Bad string; addr={hex(addr)} limit={limit} ea={hex(ea)} ret={ret}')

        return ret

    def decode_uleb128(self, readHead: int) -> Tuple[int, int]:
        """Read a Uleb128 value.
        Args:
            buffer: The data source.
            readHead: The initial offset to read from.
        Returns:
            A tuple containing the result and the new read head.
        """

        value = 0
        shift = 0

        while True:

            byte = self.get_at(readHead, 1)

            value |= (byte & 0x7f) << shift

            readHead += 1
            shift += 7

            if (byte & 0x80) == 0:
                break

        return (value, readHead)

    def _load_type(self):
        type = self.arch_struct.cputype

        if type & 0xF0000000 != 0:
            if type & 0xF == 0x7:
                return CPUType.X86_64
            elif type & 0xF == 0xC:
                return CPUType.ARM64
        else:
            if type & 0xF == 0x7:
                return CPUType.X86
            elif type & 0xF == 0xC:
                return CPUType.ARM

        raise ValueError(f'Unknown CPU Type ({hex(self.arch_struct.cputype)}) ({self.arch_struct})')

    def _load_subtype(self):
        type = self.arch_struct.cpusubtype

        if type == 3:
            return CPUSubType.X86_64_ALL
        elif type == 8:
            return CPUSubType.X86_64_H
        elif type == 9:
            return CPUSubType.ARMV7
        elif type == 11:
            return CPUSubType.ARMV7S
        elif type == 0:
            return CPUSubType.ARM64_ALL
        elif type == 1:
            return CPUSubType.ARM64_V8

        raise ValueError(f'Unknown CPU SubType ({hex(type)})')
